MODULE Program1;

IMPORT
  SYSTEM, Windows, OpenGL;

CONST
  WM_CLOSE*           = 010H;
  WM_DESTROY*         = 002H;
  WM_SIZE*            = 005H;
  WM_QUIT*            = 012H;
  WM_PAINT*           = 00FH;
  PM_REMOVE           = 1;

VAR
  defaultBrush: Windows.HBrush;
  DC: Windows.HDC;
  RC: Windows.HGLRC;
  msg : Windows.Msg;

PROCEDURE CreatePixelFormat(ColorBits, DepthBits, StencilBits: CHAR): BOOLEAN;
VAR
  Pfd: Windows.PIXELFORMATDESCRIPTOR;
  PixelFormat: LONGINT;
BEGIN
  Pfd.nSize        := SIZE(Windows.PIXELFORMATDESCRIPTOR);
  Pfd.nVersion     := 1;

  Pfd.dwFlags      := Windows.PFD_DRAW_TO_WINDOW + Windows.PFD_SUPPORT_OPENGL + Windows.PFD_DOUBLEBUFFER;
  Pfd.iPixelType   := Windows.PFD_TYPE_RGBA;

  Pfd.cColorBits   := ColorBits;
  Pfd.cDepthBits   := DepthBits;
  Pfd.cStencilBits := StencilBits;

  Pfd.iLayerType   := Windows.PFD_MAIN_PLANE;

  PixelFormat      := Windows.ChoosePixelFormat(DC,Pfd);
  IF Windows.SetPixelFormat(DC,PixelFormat,Pfd)=0 THEN RETURN FALSE; END;
  RETURN TRUE;
END CreatePixelFormat;

PROCEDURE [WINAPI] WndProc* (wnd: Windows.HWND; msg: LONGINT; wParam: Windows.WParam; lParam: Windows.LParam): Windows.LResult;
BEGIN
  CASE msg OF
    WM_CLOSE:
          Windows.DeleteObject(defaultBrush);
          Windows.PostQuitMessage(0);
          RETURN 0;
  | WM_SIZE:
    	  ReSizeGLScene(SHORT(lParam),SHORT(ASH(lParam,-16)));
          RETURN 0;
  (*
  | WM_PAINT:
          Windows.LineTo(DC,0,0);
          Windows.LineTo(DC,100,100); *)
  ELSE
    RETURN Windows.DefWindowProc(wnd, msg, wParam, lParam)
  END;
  RETURN 0;
END WndProc;

PROCEDURE OpenWindow;
VAR
  Style: SET;
  res  : LONGINT;
  rect : Windows.TRect;
  ClassName : ARRAY 32 OF CHAR;
  class     : Windows.WndClassEx;
  Handle  : Windows.HWND;
BEGIN
  ClassName           :='FORM';
  defaultBrush        := Windows.CreateSolidBrush(Windows.COLOR_SCROLLBAR);
  class.cbSize        := SIZE(Windows.WndClassEx);
  class.style         := Windows.CS_VREDRAW + Windows.CS_HREDRAW;
  class.lpfnWndProc   := WndProc;
  class.cbClsExtra    := 0;
  class.cbWndExtra    := 0;
  class.hInstance     := Windows.hInstance;
  class.hIcon         := Windows.LoadIcon(Windows.hInstance,Windows.IDI_APPLICATION);
  class.hCursor       := Windows.LoadCursor(0,SYSTEM.VAL(Windows.LPSTR,Windows.IDC_ARROW));
  class.hbrBackground := defaultBrush;
  class.lpszMenuName  := Windows.NULL;
  class.lpszClassName := SYSTEM.ADR(ClassName);
  class.hIconSm       := Windows.NULL;
  Windows.RegisterClassEx(class);

  Windows.GetWindowRect(Windows.GetDesktopWindow(),rect);

  Style:=Windows.WS_OVERLAPPEDWINDOW;
  Handle := Windows.CreateWindowEx(
     Windows.WS_EX_APPWINDOW,
     ClassName, 'Caption', Style,
     ((rect.right-640) DIV 2), ((rect.bottom-480) DIV 2),
     640, 480,
     0, 0, Windows.hInstance, 0
  );

  DC := Windows.GetDC(Handle);

  IF CreatePixelFormat(24X, 32X, 0X) THEN
    Windows.wglMakeCurrent(DC,Windows.wglCreateContext(DC));
  END;

  Windows.ShowWindow(Handle, Windows.SW_SHOW);
  Windows.UpdateWindow(Handle);

  OpenGL.glViewport(0, 0, 640, 480);
  OpenGL.glMatrixMode(OpenGL.GL_PROJECTION);
  OpenGL.glLoadIdentity();
  OpenGL.gluPerspective(45.0,640/480,0.1,100.0);
  OpenGL.glMatrixMode(OpenGL.GL_MODELVIEW);
  OpenGL.glLoadIdentity;

END OpenWindow;

PROCEDURE ReSizeGLScene(Width: OpenGL.GLsizei; Height: OpenGL.GLsizei);
BEGIN
  IF (Height=0) THEN Height:=1; END;
  OpenGL.glViewport(0, 0, Width, Height);
  OpenGL.glMatrixMode(OpenGL.GL_PROJECTION);
  OpenGL.glLoadIdentity();
  OpenGL.gluPerspective(45.0,Width/Height,0.1,100.0);
  OpenGL.glMatrixMode(OpenGL.GL_MODELVIEW);
  OpenGL.glLoadIdentity;
END ReSizeGLScene;

PROCEDURE DrawScene();
BEGIN
  OpenGL.glClear(OpenGL.GL_COLOR_BUFFER_BIT + OpenGL.GL_DEPTH_BUFFER_BIT);
  OpenGL.glLoadIdentity();
  	
  OpenGL.glTranslatef(-1.5,0.0,-6.0);
  OpenGL.glBegin(OpenGL.GL_TRIANGLES);
    OpenGL.glVertex3f( 0.0, 1.0, 0.0);
    OpenGL.glVertex3f(-1.0,-1.0, 0.0);
    OpenGL.glVertex3f( 1.0,-1.0, 0.0);
  OpenGL.glEnd();

  OpenGL.glTranslatef(3,0.0,0.0);
  OpenGL.glBegin(OpenGL.GL_QUADS);
    OpenGL.glVertex3f(-1.0, 1.0, 0.0);
    OpenGL.glVertex3f( 1.0, 1.0, 0.0);
    OpenGL.glVertex3f( 1.0,-1.0, 0.0);
    OpenGL.glVertex3f(-1.0,-1.0, 0.0);
  OpenGL.glEnd();
END DrawScene;

PROCEDURE MainLoop;
VAR
  F: BOOLEAN;
BEGIN
  F:=TRUE;
  WHILE (Windows.GetMessage(msg,0,0,0) # 0) & F DO
    Windows.TranslateMessage(msg);
    Windows.DispatchMessage(msg);

    DrawScene();
	
    Windows.SwapBuffers(DC);
    IF Windows.GetAsyncKeyState(27) # 0 THEN F:=FALSE; END;
  END;
  Windows.UnregisterClass('FORM',Windows.hInstance);
  Windows.Shutdown(0);
  Windows.ExitProcess(0);
END MainLoop;

BEGIN
  OpenWindow;

  OpenGL.glShadeModel(OpenGL.GL_SMOOTH);			
  OpenGL.glClearColor(0.0, 0.0, 0.0, 0.5);	  	
  OpenGL.glClearDepth(1.0);				
  OpenGL.glEnable(OpenGL.GL_DEPTH_TEST);			
  OpenGL.glDepthFunc(OpenGL.GL_LEQUAL);

  MainLoop;
END Program1.

